package com.qmth.distributed.print.aiDataStatistics;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class SimHash {

    private final static Logger log = LoggerFactory.getLogger(SimHash.class);

    public static final int DEFAULT_DISTANCE_THRESHOLD = 3;

    public static final int BIT_COUNT = 64;

    public static BigInteger hash(String text) {
        // 定义特征向量/数组
        int[] weight = new int[BIT_COUNT];
        Arrays.fill(weight, 0);

        // 将文本去掉格式后,分词
        List<String> tokens = Segmenter.processByIk(text);
        if (log.isTraceEnabled()) {
            log.trace("分词结果：" + StringUtils.join(tokens, " | "));
        }

        for (String token : tokens) {
            try {
                /*
                    将每一个分词hash为一组固定长度的数列.比如 64bit 的一个整数
                    建立一个长度为64的整数数组(假设要生成64位的数字指纹,也可以是其它数字)
                    对每一个分词hash后的数列进行判断,如果是1000...1,那么数组的第一位和末尾一位加1
                    中间的62位减一,也就是说,逢1加1,逢0减1.一直到把所有的分词hash数列全部判断完毕
                    这里是计算整个文档的所有特征的向量和
                    这里实际使用中需要 +- 权重，而不是简单的 +1/-1
                */

                // BigInteger hash = BigInteger.valueOf(MurmurHash.hash64(token.getBytes()));
                BigInteger hash = StringHash.hash(token, BIT_COUNT);
                for (int i = 0; i < BIT_COUNT; i++) {
                    if (hash.testBit(i)) {
                        weight[i] += 1;
                    } else {
                        weight[i] -= 1;
                    }
                }
            } catch (Exception e) {
                log.error(e.getMessage());
                continue;
            }
        }

        BigInteger fingerprint = BigInteger.valueOf(0L);
        for (int i = 0; i < BIT_COUNT; i++) {
            // 最后对数组进行判断,大于0的记为1,小于等于0的记为0,得到一个 64bit 的数字指纹/签名
            if (weight[i] > 0) {
                fingerprint = fingerprint.setBit(i);
            }
        }
        return fingerprint;
    }

    /**
     * 海明距离
     * 注：较适用于长文本匹配
     */
    public static int hmDistance(BigInteger value1, BigInteger value2) {
        int distance = 0;
        for (int i = 0; i < BIT_COUNT; i++) {
            if (!value1.shiftRight(i).and(BigInteger.ONE).equals(value2.shiftRight(i).and(BigInteger.ONE))) {
                distance++;
            }
        }
        return distance;
    }

    /**
     * 字符串编辑距离
     * 注：对于前缀匹配更友好，越往前面匹配成功权重更大
     * 可用在单词的匹配上，不推荐句子匹配，如果关键字在句子的后面部分，相似度会急剧下降，甚至为0
     */
    public static double jaroDistance(String text1, String text2) {
        return StringUtils.getJaroWinklerDistance(text1, text2);
    }

    public static boolean isDuplicate(BigInteger value1, BigInteger value2) {
        return hmDistance(value1, value2) <= DEFAULT_DISTANCE_THRESHOLD;
    }

    public static boolean isDuplicate(String text1, String text2) {
        BigInteger value1 = hash(StringUtils.trimToEmpty(text1));
        BigInteger value2 = hash(StringUtils.trimToEmpty(text2));
        return hmDistance(value1, value2) <= DEFAULT_DISTANCE_THRESHOLD;
    }

    public static List<BigInteger> groupByDistance(BigInteger value, int distance) {
        // 分成几组来检查
        int group = BIT_COUNT / (distance + 1);

        List<BigInteger> result = new ArrayList<>();
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < value.bitLength(); i++) {
            // 当且仅当设置了指定的位时，返回true
            if (value.testBit(i)) {
                buffer.append("1");
            } else {
                buffer.append("0");
            }

            if ((i + 1) % group == 0) {
                // 将二进制转为BigInteger
                BigInteger eachValue = new BigInteger(buffer.toString(), 2);
                buffer.delete(0, buffer.length());
                result.add(eachValue);
            }
        }

        return result;
    }

}
